/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <avr/eeprom.h>

#include "state.h"

#define BASE (0)
#define PERIMETER 50
#define VALUESIZE (sizeof (uint32_t))
#define CHECKSIZE (sizeof (uint8_t))
#define RINGSIZE (PERIMETER * (VALUESIZE + CHECKSIZE))
#define VALUE(i, j) (BASE + RINGSIZE * (i) + VALUESIZE * (j))
#define CHECKSUM(i, j) (BASE + RINGSIZE * (i) + VALUESIZE * PERIMETER + CHECKSIZE * (j))

#define POLYNOMIAL (0xd5)

static uint8_t cursors[4];
static uint32_t values[4];

static double oldspeed;

static uint8_t crc8(uint8_t const message[], int n)
{
    uint8_t remainder = 0;
    int i, j;
    
    for (i = 0 ; i < n ; i += 1) {
	/* Bring the next byte into the remainder. */
	
	remainder ^= message[i];

	/* Perform modulo-2 division, a bit at a time. */
	
	for (j = 8 ; j > 0 ; j -= 1) {
	    if (remainder & (1 << 7)) {
		remainder = (remainder << 1) ^ POLYNOMIAL;
	    } else {
		remainder = (remainder << 1);
	    }
	}
    }

    return (remainder);
}

void load_odometers (void)
{
    uint32_t value;
    uint8_t check;
    int i, j;

#if 0
    /* Reset the main odometer. */
    
    for (j = 0 ; j < PERIMETER ; j += 1) {
	eeprom_write_dword((uint32_t *)VALUE(3, j), 0xffffffff);
	eeprom_write_byte((uint8_t *)CHECKSUM(3, j), 0xff);
    }
#endif
    
    for (i = 0 ; i < 4 ; i += 1) {
	/* Find the largest valid value in the ring. */
	
	for (j = 0 ; j < PERIMETER ; j += 1) {
	    value = eeprom_read_dword((uint32_t *)VALUE(i, j));
	    check = eeprom_read_byte((uint8_t *)CHECKSUM(i, j));

	    if (crc8((uint8_t *)&value, 4) == check &&
		value > values[i]) {
		values[i] = value;
		cursors[i] = (j + 1) % PERIMETER;
	    }
	}
	    
	distance.odometers[i] = (float)values[i] / 10;
    }	
}

void reset_odometer (int i)
{
    int j;

    /* Reset the whole ring. */
    
    for (j = 0 ; j < PERIMETER ; j += 1) {
	eeprom_write_dword((uint32_t *)VALUE(i, j), 0xffffffff);
	eeprom_write_byte((uint8_t *)CHECKSUM(i, j), 0xff);
    }

    /* And the runtime values. */
    
    values[i] = 0;
    cursors[i] = 0;

    distance.odometers[i] = 0;
}

void update_odometers (void)
{
    double mean;
    int i, j;

    mean = 0.5 * (oldspeed + gps.speed);
    
    for (i = 0 ; i < 4 ; i += 1) {
	uint32_t value;
	
	if (i > 2 || distance.running[i]) {
	    /* Increment running odometers. */
	    
	    distance.odometers[i] += gps.delta / 3600.0 * mean;

	    /* If value has made another 0.1km step. */
	    
	    value = (uint32_t)(distance.odometers[i] * 10);

	    if (value != values[i]) {
		j = cursors[i];

		/* Update the runtime values. */
		
		values[i] = value;
		cursors[i] = (j + 1) % PERIMETER;

		/* And the ring. */
		
		eeprom_write_dword((uint32_t *)VALUE(i, j), value);
		eeprom_write_byte((uint8_t *)CHECKSUM(i, j),
				  crc8((uint8_t *) &value, 4));
	    }
	}
    }

    oldspeed = gps.speed;
}
